import cern.jet.random.engine.MersenneTwister64;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

public class MinFitnessOfNs extends LikelihoodEstimator{
	protected static final double MIN_ESTIMATED_TIME_INTERVAL = 50.0;
	protected static final double MIN_SPANNING_TIME_4_ONE_PHASE = 0.001;
	protected static final double TIME_INTERVAL_PARA = 0.01;
	protected static final int MAX_RAND_SAMPLING_CIRCLE_COUNT = 10;
	protected static final double PRECISION_THRESHOLD = 1E-4;
	protected static final int NUM_BASIC_ITERATIONS = 50000;

	protected double minChangeOfNPerPhase;
	protected double maxChangeOfNPerPhase;
	protected double minRatioOfNOverN0;
	protected double maxRatioOfNOverN0;
	protected double gammaTimeLUS;
	protected double gammaPopSizeLUS;

	protected double timeUpperBound;

	protected int numOfOmitStartSfs; // 0-based, inclusive
    protected int numOfOmitEndSfs;   // 0-based, inclusive
    protected int numOfCollapsedEndSfs;
    
    protected boolean useFoldedSFS = false; // Haipeng

    protected boolean forbidInstantaneousGrowth;
    protected boolean forbidExponentialGrowth;
    protected boolean forbidInstantaneousDecline;
    protected boolean forbidExponentialDecline;

    protected double mutRatePerGenome;
    protected int[] obsSFS;

    protected MersenneTwister64 random;

    public MinFitnessOfNs() {
    }

    protected double[] getRandomSelectedStartTimes(int numOfParameters, double[] initialTimePoints, double range) {

        double[] startTimes = new double[numOfParameters];
        startTimes[0] = 0;

        double minTime = MIN_SPANNING_TIME_4_ONE_PHASE;
        double maxTime = timeUpperBound - MIN_SPANNING_TIME_4_ONE_PHASE;

        for (int i = 1; i < numOfParameters; i ++) {
            double initialPoint = initialTimePoints[i];
            startTimes[i] = sampleBounded(initialPoint, range, minTime, maxTime);
        }
        QSort.sort(startTimes);

        return startTimes;
    }

    protected boolean examineIntervalOfStartTimes(double[] startTimes, double N0) {
        boolean feasible = true;
        for (int i = 1; i < startTimes.length; i++) {
            if ((startTimes[i] - startTimes[i - 1]) * 2 * N0 < MIN_ESTIMATED_TIME_INTERVAL) {
                feasible = false;
                break;
            }
        }
        return feasible;
    }

    protected boolean examineFeasibleOfStartTimes(double[] startTimes, int numOfTimeIntervals) {
        boolean feasible = true;

        if (numOfTimeIntervals > 1) {
            if (startTimes[1] - startTimes[0] < MIN_SPANNING_TIME_4_ONE_PHASE) {
                feasible = false;
            } else {
                for (int i = 2; i < numOfTimeIntervals; i ++) {
                    if (startTimes[i] - startTimes[i - 1] < TIME_INTERVAL_PARA) {
                        feasible = false;
                        break;
                    }
                }
            }
        }
        return feasible;
    }

    protected boolean examineFeasiblePopulation(int numOfIntervals, double[] ratiosOfNiOverN0, int[] changeTypes) {
        boolean feasible = true;
        for (int i = 1; i < numOfIntervals; i ++) {
            double changeRatio = ratiosOfNiOverN0[i] / ratiosOfNiOverN0[i - 1];
            if (changeRatio < minChangeOfNPerPhase || changeRatio > maxChangeOfNPerPhase) {
                feasible = false;
                break;
            }
        }
        if (feasible && forbidInstantaneousGrowth) {
            for (int i = 0; i < numOfIntervals - 1; i++) {
                if (changeTypes[i] == PopSizeModelExponentialLinearInstant.CHANGE_TYPE_INSTANT) {
                    if (ratiosOfNiOverN0[i] > ratiosOfNiOverN0[i + 1]) {
                        feasible = false;
                        break;
                    }
                }
            }
        }

        if (feasible && forbidInstantaneousDecline) {
            for (int i = 0; i < numOfIntervals - 1; i++) {
                if (changeTypes[i] == PopSizeModelExponentialLinearInstant.CHANGE_TYPE_INSTANT) {
                    if (ratiosOfNiOverN0[i] < ratiosOfNiOverN0[i + 1]) {
                        feasible = false;
                        break;
                    }
                }
            }
        }

        if (feasible && forbidExponentialGrowth) {
            for (int i = 0; i < numOfIntervals - 1; i++) {
                if (changeTypes[i] == PopSizeModelExponentialLinearInstant.CHANGE_TYPE_EXPONENTIAL) {
                    if (ratiosOfNiOverN0[i] > ratiosOfNiOverN0[i + 1]) {
                        feasible = false;
                        break;
                    }
                }
            }
        }

        if (feasible && forbidExponentialDecline) {
            for (int i = 0; i < numOfIntervals - 1; i++) {
                if (changeTypes[i] == PopSizeModelExponentialLinearInstant.CHANGE_TYPE_EXPONENTIAL) {
                    if (ratiosOfNiOverN0[i] < ratiosOfNiOverN0[i + 1]) {
                        feasible = false;
                        break;
                    }
                }
            }
        }
        return feasible;
    }

    public Result4SinglePopHistory optimize(int numOfTimeIntervals, TableForLinearModel[] tables, List<int[]> obsSFSs, int usedBoundary) {
        double log10MinRatioOfNOverN0 = Math.log10(minRatioOfNOverN0);
        double log10MaxRatioOfNOverN0 = Math.log10(maxRatioOfNOverN0);
        List<double[]> logSFSs       = new ArrayList<>();
        List<double[]> logFoldedSFSs = new ArrayList<>(); // Haipeng
        
        for (int i = 0; i < usedBoundary; i++) {
            logSFSs.add(initialLogSFSFactorial(obsSFSs.get(i)));
            
            // Haipeng
            int[] oneSFS = obsSFSs.get(i);
            int n = oneSFS.length + 1;
            
            int[] foldedSFS = new int[n / 2];
            if (n % 2 == 0) { // an even number, 4, 6, 8, 10, ....
            	for (int j = 0; j < foldedSFS.length - 1; j ++) {
            		foldedSFS[j] = oneSFS[j] + oneSFS[n - j - 2];
            	}
            	foldedSFS[foldedSFS.length - 1] = oneSFS[foldedSFS.length - 1];
            }
            else { // an odd number, 5, 7, 9, 11, ....
            	for (int j = 0; j < foldedSFS.length; j ++) {
            		foldedSFS[j] = oneSFS[j] + oneSFS[n - j - 2];
            	}
            }
            
            logFoldedSFSs.add(initialLogSFSFactorial(foldedSFS));
        }
        
        int[] collapsedObsMuts = null;
        double[] logFactorialOfCollapsedObsMuts = null;
        if (numOfCollapsedEndSfs > 0) {
        	int numOfDataset = obsSFSs.size();
        	collapsedObsMuts = new int[numOfDataset];
        	logFactorialOfCollapsedObsMuts = new double[numOfDataset];
        	
        	for (int dataIndex = 0; dataIndex < numOfDataset; dataIndex ++) {
        		collapsedObsMuts[ dataIndex ] = getCollapsedObsMuts(obsSFSs.get(dataIndex), numOfCollapsedEndSfs);
            	logFactorialOfCollapsedObsMuts[ dataIndex ] = getLogFactorial(collapsedObsMuts[ dataIndex ]);
        	}
        }

        int[] currentChangeTypes = new int[numOfTimeIntervals];
        double[] currentStartTimes = new double[numOfTimeIntervals];
        double[] currentRatios = new double[numOfTimeIntervals];

        int maxIterations = NUM_BASIC_ITERATIONS * (numOfTimeIntervals - 1) + 1;

        Result4SinglePopHistory bestResult = null;
        double minFitness = Double.MAX_VALUE;

        int numOfChangeTypes;
        boolean onlyInstantChange = forbidExponentialDecline && forbidExponentialGrowth;
        boolean onlyExponentialChange = forbidInstantaneousDecline && forbidInstantaneousGrowth;
        if (onlyInstantChange || onlyExponentialChange) {
            numOfChangeTypes = getNumOfChangeTypes();
        } else {
            numOfChangeTypes = getNumOfChangeTypes(numOfTimeIntervals);
        }

        for (int indexOfChangeTypes = 0; indexOfChangeTypes < numOfChangeTypes; indexOfChangeTypes++) {
            int[] count = new int[1];
            if (onlyInstantChange) {
                setChangeTypes(numOfTimeIntervals, currentChangeTypes, 1);
            } else if (onlyExponentialChange) {
                setChangeTypes(numOfTimeIntervals, currentChangeTypes, 3);
            } else {
                setChangeTypes(indexOfChangeTypes, count, 0, currentChangeTypes, numOfTimeIntervals);
            }

            double rPopSize = 1.0;
            double qPopSize = Math.pow(2.0, -1.0 / (1 + (numOfTimeIntervals - 1) * gammaPopSizeLUS));

            double rStartTimes = timeUpperBound;
            double qStartTimes = Math.pow(2.0, -1.0 / (1 + (numOfTimeIntervals - 1) * gammaTimeLUS));

            double[] bestStartTimesGivenChangeTypes = new double[numOfTimeIntervals];
            double[] bestRatiosGivenChangeTypes = new double[numOfTimeIntervals];

            for (int i = 1; i < numOfTimeIntervals; i++) {
                bestRatiosGivenChangeTypes[i] = 1.0d;
            }

            Result4SinglePopHistory bestResultGivenChangeTypes = new Result4SinglePopHistory();
            double minFitnessGivenChangeTypes = Double.MAX_VALUE;

            for (int iteration = 1; iteration <= maxIterations; iteration++) {
                int circleCount = 0;
                boolean feasible = true;
                currentRatios[0] = 1.0d;
                do {
                    for (int i = 1; i < numOfTimeIntervals; i++) {
                        double rand = sampleBounded(Math.log10(bestRatiosGivenChangeTypes[i]), rPopSize, log10MinRatioOfNOverN0, log10MaxRatioOfNOverN0);
                        currentRatios[i] = Math.pow(10, rand);
                    }
                    feasible = examineFeasiblePopulation(numOfTimeIntervals, currentRatios, currentChangeTypes);
                    circleCount++;
                } while (circleCount < MAX_RAND_SAMPLING_CIRCLE_COUNT && (!feasible));

                if (!feasible) {
                    continue;
                }

                circleCount = 0;
                currentStartTimes[0] = 0.0d;
                do {
                    currentStartTimes = getRandomSelectedStartTimes(numOfTimeIntervals, bestStartTimesGivenChangeTypes, rStartTimes);
                    feasible = examineFeasibleOfStartTimes(currentStartTimes, numOfTimeIntervals);
                    circleCount++;
                } while (circleCount < MAX_RAND_SAMPLING_CIRCLE_COUNT && (!feasible));

                if (!feasible) {
                    continue;
                }

                PopSizeModelExponentialLinearInstant popModel = new PopSizeModelExponentialLinearInstant();
                popModel.setParameters(currentStartTimes, currentRatios, currentChangeTypes.clone());

                PopSizeModelInstantChange[] convertedModels = new PopSizeModelInstantChange[usedBoundary];
                for (int i = 0; i < usedBoundary; i++) {
                    double[] tabulatedTimes = tables[i].startTimeOfDetailedIntervals;
                    convertedModels[i] = popModel.convert(tabulatedTimes);
                }

                double N0 = 0;
                double totalFitness = 0;
                double ratioOfOmitStart = (double) numOfOmitStartSfs / (obsSFSs.get(0).length + 1);
                double ratioOfOmitLast = (double) numOfOmitEndSfs / (obsSFSs.get(0).length + 1);
                for (int i = 0; i < usedBoundary; i++) {
                    TableForLinearModel table = tables[i];
                    PopSizeModelInstantChange convertedModel = convertedModels[i];
                    int[] obsSFS = obsSFSs.get(i);
                    double[] logSFS = logSFSs.get(i);
                    double[] logFoldedSFS = logFoldedSFSs.get(i); // Haipeng
                    
                    int omitStart = (int) Math.round(ratioOfOmitStart * (obsSFS.length + 1));
                    int omitLast  = (int) Math.round(ratioOfOmitLast  * (obsSFS.length + 1));
                    double[] expectedBranchLengthsIn2N0 = table.getExpectedBranchLengths(convertedModel);

                    // omitStart: 0-based, inclusive
                    // endIndex: 0-based, inclusive
                    double obsMutSum = sum(obsSFS, omitStart, obsSFS.length - omitLast - 1);
                    double expectedBranchLengthsIn2N0Sum = sum(table.getExpectedBranchLengths(convertedModel), omitStart, obsSFS.length - omitLast - 1);
                    // Haipeng: when considering the folded SFS, we will have only numOfOmitStartSfs
                    if (useFoldedSFS) {
                    	obsMutSum = sum(obsSFS, omitStart, obsSFS.length - omitStart - 1);
                    	expectedBranchLengthsIn2N0Sum = sum(expectedBranchLengthsIn2N0, omitStart, obsSFS.length - omitStart - 1);
                    }

                    double thetaScaledIn2N0 = obsMutSum / expectedBranchLengthsIn2N0Sum;
                    double weightedN0 = thetaScaledIn2N0 / mutRatePerGenome * 0.5;

                    N0 += weightedN0;
                    
                    // Haipeng
                    if (useFoldedSFS) {
                    	// initialSize: 1-based, inclusive
                        // endSize: 1-based, inclusive
                        double localFitness_Folded = estimateFitness_Folded(thetaScaledIn2N0, expectedBranchLengthsIn2N0, obsSFS, logFoldedSFS, 1 + omitStart);
                        totalFitness += localFitness_Folded;
                    }
                    else {
                    	if (numOfCollapsedEndSfs == 0) {
                    		// initialSize: 1-based, inclusive
                            // endSize: 1-based, inclusive
                            double localFitness = estimateFitness(thetaScaledIn2N0, expectedBranchLengthsIn2N0, obsSFS, logSFS, 1 + omitStart, obsSFS.length - omitLast);
                            totalFitness += localFitness;
                    	}
                    	else {
                    		// estimateFitness_CollapsedTail(double thetaScaledIn2N0, double[] expectedBranchLengthsIn2N0, int[] obsSFS, double[] logSFS, int initialSize, int numOfCollapsedEndSfs, int collapsedObsMuts, double logCollapsedMuts) {
                    		double localFitness = estimateFitness_CollapsedTail(thetaScaledIn2N0, expectedBranchLengthsIn2N0, obsSFS, logSFS, 1 + omitStart, numOfCollapsedEndSfs, collapsedObsMuts[i], logFactorialOfCollapsedObsMuts[i]);
                            totalFitness += localFitness;
                    	}
                    }
                }

                if (minFitnessGivenChangeTypes > totalFitness) {
                    minFitnessGivenChangeTypes = totalFitness;
                    bestResultGivenChangeTypes.N0 = N0;
                    bestResultGivenChangeTypes.fitness = totalFitness;
                    bestResultGivenChangeTypes.popSizeChangeTypes = currentChangeTypes.clone();
                    bestResultGivenChangeTypes.startTimeOfIntervals = currentStartTimes.clone();
                    bestResultGivenChangeTypes.ratiosOfNiOverN0 = currentRatios.clone();

                    double[] temp = bestStartTimesGivenChangeTypes;
                    bestStartTimesGivenChangeTypes = currentStartTimes;
                    currentStartTimes = temp;

                    temp = bestRatiosGivenChangeTypes;
                    bestRatiosGivenChangeTypes = currentRatios;
                    currentRatios = temp;
                } else {
                    rStartTimes *= qStartTimes;
                    rPopSize *= qPopSize;
                    if ((rStartTimes < PRECISION_THRESHOLD) && (rPopSize < PRECISION_THRESHOLD)) {
                        break;
                    }
                }
            }
            if (minFitness > minFitnessGivenChangeTypes) {
                minFitness = minFitnessGivenChangeTypes;
                bestResult = bestResultGivenChangeTypes;
            }
        }
        return bestResult;
    }

    public Result4SinglePopHistory optimize(int numOfTimeIntervals, TableForLinearModel table){
        double log10MinRatioOfNOverN0 = Math.log10(minRatioOfNOverN0);
        double log10MaxRatioOfNOverN0 = Math.log10(maxRatioOfNOverN0);
        double[] logSFS = initialLogSFSFactorial(obsSFS);
        
        // Haipeng
        int n = obsSFS.length + 1;
        int[] foldedSFS = new int[n / 2];
        if (n % 2 == 0) { // an even number, 4, 6, 8, 10, ....
        	for (int j = 0; j < foldedSFS.length - 1; j ++) {
        		foldedSFS[j] = obsSFS[j] + obsSFS[n - j - 2];
        	}
        	foldedSFS[foldedSFS.length - 1] = obsSFS[foldedSFS.length - 1];
        }
        else { // an odd number, 5, 7, 9, 11, ....
        	for (int j = 0; j < foldedSFS.length; j ++) {
        		foldedSFS[j] = obsSFS[j] + obsSFS[n - j - 2];
        	}
        }
        double[] logFoldedSFS = initialLogSFSFactorial(foldedSFS);
        
        int collapsedObsMuts = 0;
        double logFactorialOfCollapsedObsMuts = 0;
        if (numOfCollapsedEndSfs > 0) {
        	collapsedObsMuts = getCollapsedObsMuts(obsSFS, numOfCollapsedEndSfs);
        	logFactorialOfCollapsedObsMuts = getLogFactorial(collapsedObsMuts);
        }

        int[] currentChangeTypes = new int[numOfTimeIntervals];
        double[] currentStartTimes = new double[numOfTimeIntervals];
        double[] currentRatios = new double[numOfTimeIntervals];

        int maxIterations = NUM_BASIC_ITERATIONS * (numOfTimeIntervals - 1) + 1;

        Result4SinglePopHistory bestResult = null;
        double minFitness = Double.MAX_VALUE;

        int numOfChangeTypes;
        boolean onlyInstantChange = forbidExponentialDecline && forbidExponentialGrowth;
        boolean onlyExponentialChange = forbidInstantaneousDecline && forbidInstantaneousGrowth;
        if (onlyInstantChange || onlyExponentialChange) {
            numOfChangeTypes = getNumOfChangeTypes();
        } else {
            numOfChangeTypes = getNumOfChangeTypes(numOfTimeIntervals);
        }

        for (int indexOfChangeTypes = 0; indexOfChangeTypes < numOfChangeTypes; indexOfChangeTypes++) {
            int[] count = new int[1];
            if (onlyInstantChange) {
                setChangeTypes(numOfTimeIntervals, currentChangeTypes, 1);
            } else if (onlyExponentialChange) {
                setChangeTypes(numOfTimeIntervals, currentChangeTypes, 3);
            } else {
                setChangeTypes(indexOfChangeTypes, count, 0, currentChangeTypes, numOfTimeIntervals);
            }

            double rPopSize = 1.0;
            double qPopSize = Math.pow(2.0, -1.0 / (1 + (numOfTimeIntervals - 1) * gammaPopSizeLUS));
            double rStartTimes = timeUpperBound;
            double qStartTimes = Math.pow(2.0, -1.0 / (1 + (numOfTimeIntervals - 1) * gammaPopSizeLUS));

            double[] bestStartTimesGivenChangeTypes = new double[numOfTimeIntervals];
            double[] bestRatiosGivenChangeTypes = new double[numOfTimeIntervals];

            for (int i = 1; i < numOfTimeIntervals; i++) {
                bestRatiosGivenChangeTypes[i] = 1.0d;
            }


            Result4SinglePopHistory bestResultGivenChangeTypes = new Result4SinglePopHistory();
            double minFitnessGivenChangeTypes = Double.MAX_VALUE;

            for (int iteration = 1; iteration <= maxIterations; iteration++) {
                int circleCount = 0;
                boolean feasible = true;
                currentRatios[0] = 1.0d;
                do {
                    for (int i = 1; i < numOfTimeIntervals; i++) {
                        double rand = sampleBounded(Math.log10(bestRatiosGivenChangeTypes[i]), rPopSize, log10MinRatioOfNOverN0, log10MaxRatioOfNOverN0);
                        currentRatios[i] = Math.pow(10, rand);
                    }
                    feasible = examineFeasiblePopulation(numOfTimeIntervals, currentRatios, currentChangeTypes);
                    circleCount++;
                } while (circleCount < MAX_RAND_SAMPLING_CIRCLE_COUNT && (!feasible));

                if (!feasible) {
                    continue;
                }

                circleCount = 0;
                currentStartTimes[0] = 0.0d;
                do {
                    currentStartTimes = getRandomSelectedStartTimes(numOfTimeIntervals, bestStartTimesGivenChangeTypes, rStartTimes);
                    feasible = examineFeasibleOfStartTimes(currentStartTimes, numOfTimeIntervals);
                    circleCount++;
                } while (circleCount < MAX_RAND_SAMPLING_CIRCLE_COUNT && (!feasible));

                if (!feasible) {
                    continue;
                }
                PopSizeModelExponentialLinearInstant popModel = new PopSizeModelExponentialLinearInstant();
                popModel.setParameters(currentStartTimes, currentRatios, currentChangeTypes.clone());

                double[] tabulatedTimes = table.startTimeOfDetailedIntervals;
                PopSizeModelInstantChange convertedModel = popModel.convert(tabulatedTimes);
                double[] expectedBranchLengthsIn2N0 = table.getExpectedBranchLengths(convertedModel);

                // numOfOmitStartSfs: 0-based, inclusive
                // endIndex: 0-based, inclusive
                double obsMutSum = sum(obsSFS, numOfOmitStartSfs, obsSFS.length - numOfOmitEndSfs - 1);
                double expectedBranchLengthsIn2N0Sum = sum(expectedBranchLengthsIn2N0, numOfOmitStartSfs, obsSFS.length - numOfOmitEndSfs - 1);
                // Haipeng: when considering the folded SFS, we will have only numOfOmitStartSfs
                if (useFoldedSFS) {
                	obsMutSum = sum(obsSFS, numOfOmitStartSfs, obsSFS.length - numOfOmitStartSfs - 1);
                	expectedBranchLengthsIn2N0Sum = sum(expectedBranchLengthsIn2N0, numOfOmitStartSfs, obsSFS.length - numOfOmitStartSfs - 1);
                }

                double thetaScaledIn2N0 = obsMutSum / expectedBranchLengthsIn2N0Sum;
                double N0 = thetaScaledIn2N0 / mutRatePerGenome * 0.5;
                feasible = examineIntervalOfStartTimes(currentStartTimes, N0);
                if (!feasible) {
                    continue;
                }

                // Haipeng
                double localFitness = 0;
                if (useFoldedSFS) {
                	// initialSize: 1-based, inclusive
                	localFitness = estimateFitness_Folded(thetaScaledIn2N0, expectedBranchLengthsIn2N0, obsSFS, logFoldedSFS, 1 + numOfOmitStartSfs);
                }
                else {
                	
                	if (numOfCollapsedEndSfs == 0) {
                		// initialSize: 1-based, inclusive
                        // endSize: 1-based, inclusive
                        localFitness = estimateFitness(thetaScaledIn2N0, expectedBranchLengthsIn2N0, obsSFS, logSFS, 1 + numOfOmitStartSfs, obsSFS.length - numOfOmitEndSfs);
                	}
                	else {
                		// estimateFitness_CollapsedTail(double thetaScaledIn2N0, double[] expectedBranchLengthsIn2N0, int[] obsSFS, double[] logSFS, int initialSize, int numOfCollapsedEndSfs, int collapsedObsMuts, double logCollapsedMuts) {
                		localFitness = estimateFitness_CollapsedTail(thetaScaledIn2N0, expectedBranchLengthsIn2N0, obsSFS, logSFS, 1 + numOfOmitStartSfs, numOfCollapsedEndSfs, collapsedObsMuts, logFactorialOfCollapsedObsMuts);
                	}
                }
                
                if (minFitnessGivenChangeTypes > localFitness) {
                    minFitnessGivenChangeTypes = localFitness;
                    bestResultGivenChangeTypes.N0 = N0;
                    bestResultGivenChangeTypes.fitness = localFitness;
                    bestResultGivenChangeTypes.popSizeChangeTypes = currentChangeTypes.clone();
                    bestResultGivenChangeTypes.startTimeOfIntervals = currentStartTimes.clone();
                    bestResultGivenChangeTypes.ratiosOfNiOverN0 = currentRatios.clone();

                    double[] temp = bestStartTimesGivenChangeTypes;
                    bestStartTimesGivenChangeTypes = currentStartTimes;
                    currentStartTimes = temp;

                    temp = bestRatiosGivenChangeTypes;
                    bestRatiosGivenChangeTypes = currentRatios;
                    currentRatios = temp;

                } else {
                    rStartTimes *= qStartTimes;
                    rPopSize *= qPopSize;
                    if ((rStartTimes < PRECISION_THRESHOLD) && (rPopSize < PRECISION_THRESHOLD)) {
                        break;
                    }
                }
            }
            if (minFitness > minFitnessGivenChangeTypes) {
                minFitness = minFitnessGivenChangeTypes;
                bestResult = bestResultGivenChangeTypes;
            }
        }
        
        return bestResult;
    }

    protected int getNumOfChangeTypes() {
        return 1;
    }

    protected int getNumOfChangeTypes(int numOfParameters) {
        int numOfChangeTypes = 1;
        for (int i = 0; i < numOfParameters - 1; i ++) {
            numOfChangeTypes *= 2;
        }

        return numOfChangeTypes;
    }

    protected void setChangeTypes(int numOfTimeIntervals, int[] changeTypes, int type) {
        if (type == 1) {
            changeTypes[numOfTimeIntervals - 1] = PopSizeModelExponentialLinearInstant.CHANGE_TYPE_LAST_CONSTANT;
            for (int i = 0; i < numOfTimeIntervals - 1; i++) {
                changeTypes[i] = PopSizeModelExponentialLinearInstant.CHANGE_TYPE_INSTANT;
            }
        } else {
            changeTypes[numOfTimeIntervals - 1] = PopSizeModelExponentialLinearInstant.CHANGE_TYPE_LAST_CONSTANT;
            for (int i = 0; i < numOfTimeIntervals - 1; i++) {
                changeTypes[i] = PopSizeModelExponentialLinearInstant.CHANGE_TYPE_EXPONENTIAL;
            }
        }

    }

    protected void setChangeTypes(int indexOfChangeTypes, int[] count, int layerIndex, int[] changeTypes, int numOfParameters) {
        if (layerIndex == numOfParameters - 1) {
            if (count[0] <= indexOfChangeTypes) {
                changeTypes[layerIndex] = PopSizeModelExponentialLinearInstant.CHANGE_TYPE_LAST_CONSTANT;
                count[0]++;
            }
        } else {
            if (count[0] <= indexOfChangeTypes) {
                changeTypes[layerIndex] = PopSizeModelExponentialLinearInstant.CHANGE_TYPE_INSTANT;
                setChangeTypes(indexOfChangeTypes, count, layerIndex + 1, changeTypes, numOfParameters);
            }
            if (count[0] <= indexOfChangeTypes) {
                changeTypes[layerIndex] = PopSizeModelExponentialLinearInstant.CHANGE_TYPE_EXPONENTIAL;
                setChangeTypes(indexOfChangeTypes, count, layerIndex + 1, changeTypes, numOfParameters);
            }
        }
    }

    protected double sampleBounded(double x, double range, double lowerBound, double upperBound) {
        double l = Math.max(x - range, lowerBound);
        double u = Math.min(x + range, upperBound);

        double y = random.nextDouble();
        if (u >= l) {
            y = y * (u - l) + l;
        } else {
            y = y * (l - u) + u;
        }

        return y;
    }

    public void setObsSFS(int[] obsSFS) {
        this.obsSFS = obsSFS;
    }

    // initialIndex: 0-based, inclusive
    // endIndex: 0-based, inclusive
    protected int sum(int[] values, int initialIndex, int endIndex) {
        int sumValue = 0;
        for (int i = initialIndex; i <= endIndex; i ++) {
            sumValue += values[i];
        }

        return sumValue;
    }

    // initialIndex: 0-based, inclusive
    // endIndex: 0-based, inclusive
    protected double sum(double[] values, int initialIndex, int endIndex) {
        double sumValue = 0;
        for (int i = initialIndex; i <= endIndex; i ++) {
            sumValue += values[i];
        }

        return sumValue;
    }

    // initialSize: 1-based, inclusive
    // endSize: 1-based, inclusive
    protected double estimateFitness(double thetaScaledIn2N0, double[] expectedBranchLengthsIn2N0, int[] obsSFS, double[] logSFS, int initialSize, int endSize) {
        double logL = 0;
        for (int size = initialSize; size <= endSize; size++) {
            double expectedMut = expectedBranchLengthsIn2N0[size - 1] * thetaScaledIn2N0;
            logL += logLikelihood(expectedMut, obsSFS[size - 1], logSFS[size - 1]);
        }

        return -logL;
    }
    
    // initialSize: 1-based, inclusive
    // numOfCollapsedEndSfs
    protected double estimateFitness_CollapsedTail(double thetaScaledIn2N0, double[] expectedBranchLengthsIn2N0, int[] obsSFS, double[] logSFS, int initialSize, int numOfCollapsedEndSfs, int collapsedObsMuts, double logFactorialOfCollapsedObsMuts) {
        double logL = 0;
        
        int n = obsSFS.length + 1;
        int endSize = n - 1 - numOfCollapsedEndSfs;
        for (int size = initialSize; size <= endSize; size++) {
            double expectedMut = expectedBranchLengthsIn2N0[size - 1] * thetaScaledIn2N0;
            logL += logLikelihood(expectedMut, obsSFS[size - 1], logSFS[size - 1]);
        }
        
        double collapsedExpectedMuts = 0;
        for (int size = endSize + 1; size < n; size ++) {
        	collapsedExpectedMuts += expectedBranchLengthsIn2N0[size - 1];
        }
        collapsedExpectedMuts *= thetaScaledIn2N0;
        
        logL += logLikelihood(collapsedExpectedMuts, collapsedObsMuts, logFactorialOfCollapsedObsMuts);

        return -logL;
    }
    
    // omitStart: 1-based, inclusive
    // Haipeng
    protected double estimateFitness_Folded(double thetaScaledIn2N0, double[] expectedBranchLengthsIn2N0, int[] obsSFS, double[] logFoldedSFS, int omitStart) {
        double logL = 0;
        
        int sampleSize = obsSFS.length + 1;
        int lengthOfLogFoldedSFS = logFoldedSFS.length;
        
        if (sampleSize % 2 == 0) { // an even number, 4, 6, 8, 10, ....
        	for (int foldedSize = omitStart; foldedSize < lengthOfLogFoldedSFS; foldedSize ++) {
                double expectedMut01 = expectedBranchLengthsIn2N0[foldedSize - 1]              * thetaScaledIn2N0;
                double expectedMut02 = expectedBranchLengthsIn2N0[sampleSize - foldedSize - 1] * thetaScaledIn2N0;
                
                int numOfMuts = obsSFS[foldedSize - 1] + obsSFS[sampleSize - foldedSize - 1];
                
                logL += logLikelihood(expectedMut01 + expectedMut02, numOfMuts, logFoldedSFS[foldedSize - 1]);
            }
        	
        	double expectedMut = expectedBranchLengthsIn2N0[lengthOfLogFoldedSFS - 1] * thetaScaledIn2N0;
        	int numOfMuts = obsSFS[lengthOfLogFoldedSFS - 1];
        	logL += logLikelihood(expectedMut, numOfMuts, logFoldedSFS[lengthOfLogFoldedSFS - 1]);
        }
        else { // an odd number, 5, 7, 9, 11, ....
        	for (int foldedSize = omitStart; foldedSize <= lengthOfLogFoldedSFS; foldedSize ++) {
                double expectedMut01 = expectedBranchLengthsIn2N0[foldedSize - 1]              * thetaScaledIn2N0;
                double expectedMut02 = expectedBranchLengthsIn2N0[sampleSize - foldedSize - 1] * thetaScaledIn2N0;
                
                int numOfMuts = obsSFS[foldedSize - 1] + obsSFS[sampleSize - foldedSize - 1];
                
                logL += logLikelihood(expectedMut01 + expectedMut02, numOfMuts, logFoldedSFS[foldedSize - 1]);
            }
        }

        return -logL;
    }
    
    protected int getCollapsedObsMuts(int[] obsSFS, int numOfCollapsedEndSfs) {
    	int n = obsSFS.length + 1;
        int endSize = n - 1 - numOfCollapsedEndSfs;
        
        int collapsedObsMuts = 0;
        for (int size = endSize + 1; size < n; size ++) {
        	collapsedObsMuts += obsSFS[size - 1];
        }
        
        return collapsedObsMuts;
    }
    
    protected double getLogFactorial(int value) {
    	double logFactorial = 0;
    	for (int i = 1; i <= value; i ++) {
    		logFactorial += Math.log(i);
        }

        return logFactorial;
    }

    public void setMutRatePerGenome(double mutRatePerGenome) {
        this.mutRatePerGenome = mutRatePerGenome;
    }

    public void setRandom(MersenneTwister64 random) {
        this.random = random;
    }

    public void setNumOfOmitStartSfs(int numOfOmitStartSfs) {
        this.numOfOmitStartSfs = numOfOmitStartSfs;
    }

    public void setNumOfOmitEndSfs(int numOfOmitEndSfs) {
        this.numOfOmitEndSfs = numOfOmitEndSfs;
    }
    
    public void setNumOfCollapsedEndSfs(int numOfCollapsedEndSfs) {
        this.numOfCollapsedEndSfs = numOfCollapsedEndSfs;
    }

    public void setMinChangeOfNPerPhase(double minChangeOfNPerPhase) {
        this.minChangeOfNPerPhase = minChangeOfNPerPhase;
    }

    public void setMaxChangeOfNPerPhase(double maxChangeOfNPerPhase) {
        this.maxChangeOfNPerPhase = maxChangeOfNPerPhase;
    }

    public void setMinRatioOfNOverN0(double minRatioOfNOverN0) {
        this.minRatioOfNOverN0 = minRatioOfNOverN0;
    }

    public void setMaxRatioOfNOverN0(double maxRatioOfNOverN0) {
        this.maxRatioOfNOverN0 = maxRatioOfNOverN0;
    }

    public void setGammaTimeLUS(double gammaTimeLUS) {
        this.gammaTimeLUS = gammaTimeLUS;
    }

    public void setGammaPopSizeLUS(double gammaPopSizeLUS) {
        this.gammaPopSizeLUS = gammaPopSizeLUS;
    }

    public void setTimeUpperBound(double timeUpperBound) {
        this.timeUpperBound = timeUpperBound;
    }

    public void setForbidInstantaneousGrowth(boolean forbidInstantaneousGrowth) {
        this.forbidInstantaneousGrowth = forbidInstantaneousGrowth;
    }

    public void setForbidExponentialGrowth(boolean forbidExponentialGrowth) {
        this.forbidExponentialGrowth = forbidExponentialGrowth;
    }

    public void setForbidInstantaneousDecline(boolean forbidInstantaneousDecline) {
        this.forbidInstantaneousDecline = forbidInstantaneousDecline;
    }

    public void setForbidExponentialDecline(boolean forbidExponentialDecline) {
        this.forbidExponentialDecline = forbidExponentialDecline;
    }
    
    // Haipeng
    public void setUseFoldedSFS(boolean useFoldedSFS) {
        this.useFoldedSFS = useFoldedSFS;
    }
}
